/*
 * C111A Driver - by G. Organtini 2000 
 * for RedHat 6.0+
 * 
 * based on C111 driver from Leonardi+Organtini
 * rewritten to cope with new OS standards
 *
 */

#define MODULE
/*#define _DEBUG*/

#include <naf.h>

#include <linux/module.h>
#include <linux/init.h>

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/wrapper.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/system.h>

#define LinuxVersionCode(v, p, s) (((v)<<16)+((p)<<8)+(s))
#if LINUX_VERSION_CODE < LinuxVersionCode(2,4,0)
#define WAKE_UP module_wake_up
#define INTERRUPTIBLE_SLEEP_ON module_interruptible_sleep_on
#else
#define WAKE_UP wake_up
#define INTERRUPTIBLE_SLEEP_ON interruptible_sleep_on
#endif

/* The major no. */
#define C111A_MAJOR 27
/* The IRQ line */
#define IRQ 0x09
/* The pc board address */
#define C111A_BASE_ADDRESS 0x380
/* the amount of memory to allocate */
#define C111A_BASE_EXTEND 0x1A
/* the number of devices */
#define C111A_MAX_CRATES 16
/* the delay for the cable */
#define DELAY 0x01

#define FIFOS 2
#define FIFO_SIZE 500000
#define CTRLFIFO 0
#define DATAFIFO 1

#define C111A_NAME  "C111A"
#define STARTUPSTRING "C111A CAMAC Crate Controller Module\nby G. Organtini 2000\n"
static unsigned long base = C111A_BASE_ADDRESS;
#define GEO_port  (base)
#define FNA_port  (base + 0x02)
#define HB_W      (base + 0x04)
#define LW_W      (base + 0x06)
#define HB_R      (base + 0x08)
#define LW_R      (base + 0x0A)
#define BR_R      (base + 0x0C)
#define S_mod_W   (base + 0x0E)
#define S_mod_R   (base + 0x10)
#define R_S       (base + 0x12)
#define Crate_C   (base + 0x14)
#define Crate_Z   (base + 0x16)
#define Crate_on  (base + 0x18)
#define Crate_off (base + 0x1A)

static int major = C111A_MAJOR;
static int currentDevice = 0;
static char nospace_notified = 0;

static char nDeviceOpen[C111A_MAX_CRATES+FIFOS];

/*--------- collect statistics ----------------*/

static int zero_readings;
static int empty_fifo;
static int cycling;
static int time_since_last_interrupt;
static int mean_time_between_interrupts;
static int interrupts;
static int sleep_on_counter;
static int wake_up;

/*--------- forward declarations --------------*/

static int c111a_set(unsigned int crate);
static void c111a_load_naf(int naf);
static int c111a_status();

/*--------- The fifo functions ----------------*/

/* The queue of processes waiting for data */
#if LINUX_VERSION_CODE < LinuxVersionCode(2,4,0)
static struct wait_queue * wait_on_data[FIFOS]; 
#else
static wait_queue_head_t wait_on_data[FIFOS];
#endif

static int fifo[FIFOS][FIFO_SIZE];
static int first[FIFOS];
static int last[FIFOS];

static int rtf_create(int which) {
  int i=0;
  cycling = 0;
  for (i=0; i<FIFO_SIZE; i++) {
    fifo[which][i] = 0;
  }
  first[which]=0;
  last[which]=0;
  return 0;
}

static int rtf_destroy(int which) {
  return rtf_create(which);
}

static int rtf_get(int which, char *buffer, int size) {
  int count = 0;
  int *data = kmalloc(sizeof(int), GFP_KERNEL);
#ifdef _DEBUG
  printk("C111A DEBUG: FIFO%i first=%i last=%i\n", 
	 which, first[which], last[which]);
#endif
  if (first[which] == last[which]) {
#ifdef _DEBUG
    printk("C111A DEBUG: NO DATA!\n");
#endif
    kfree(data);
    zero_readings++;
    return 0;
  }
  if (data == NULL) {
    printk("C111A NOTIFY: Unable to kmalloc in rtf_get\n");
    return 0;
  }
  while (count < size) {
    *data = fifo[which][first[which]++];
#ifdef _DEBUG
    printk("C111A DEBUG: DATA=%i\n", *data);
#endif
    memcpy(buffer, data, sizeof(int));
    buffer += sizeof(int);
    count += sizeof(int);
    if (first[which] >= last[which]) {
      //
      // May be this is an error!!!
      //
      if (first[which] > last[which]) {
	printk("C111A DEBUG: first data in FIFO beyond last data\n");
      }
      empty_fifo++;
      first[which] = 0;
      last[which] = 0;
      kfree(data);
      return count;
    }
    if (first[which]>=FIFO_SIZE) {
      printk("C111A DEBUG: End of fifo reached: Putting first to 0\n");
      first[which] = 0;
    }
  }
  kfree(data);
#ifdef _DEBUG
  printk("C111A DEBUG: %i bytes passed from fifo to buffer!\n", count);
#endif
  return count;
}

static int rtf_put(int which, int *buffer, int size) {
  /* Check if there is enough space left in the queue */

  int *k = buffer;
  int i=0;
  int spaceleft = FIFO_SIZE;
#ifdef _DEBUG
  printk("C111A DEBUG: (rtf_put) Data size: %i\n", size);
#endif
  if (last[which]==first[which]) {
    spaceleft = FIFO_SIZE;
  } else if (last[which]>first[which]) {
    spaceleft -= (last[which]-first[which]);
  } else {
    spaceleft -= (first[which]-last[which]);
    printk("C111A DEBUG: (rtf_put) mmmmhhh...something wrong? F=%i L=%i SL=%i FS=%i\n", 
	   first[which], last[which], spaceleft, FIFO_SIZE);
  }
#ifdef _DEBUG
  printk("C111A DEBUG: Putting %i bytes on fifo %i\n", size, which);
#endif
  if (spaceleft < size) {
    if (!nospace_notified) {
      printk("C111A DEBUG: No more space left in fifo...");
      nospace_notified = 1;
    }
    return 0;
  } else {
    nospace_notified = 0;
  }
  /* copy buffer in the queue */
#ifdef _DEBUG
  printk("C111A DEBUG: ============= %i:%i ============\n", size, which);
  for (i=0; i<size; i++) {
    printk("C111A DEBUG: Buffer Content %i\n", *k);
    k++;
  }
  printk("C111A DEBUG: ============== %i =================\n", 
	 size);
#endif
  k = buffer;
  for (i=0; i<size; i++) {
    fifo[which][last[which]++] = *k;
    if (last[which]>=FIFO_SIZE) {
      /* was last[which] = 1 */
      cycling++;
      printk("C111A DEBUG: fifo full - restarting from top    \n"); 
      last[which] = 0;
    }
    k++;
  }
#ifdef _DEBUG
  printk("C111A DEBUG: Waking up fifo %i (last=%i; first = %i)\n", which, last[which], first[which]);
#endif
  wake_up++;
  WAKE_UP(&wait_on_data[which]);
  return size;
};

static ssize_t read_fifo(struct file * file, char * buf, size_t count, 
			 loff_t *ppos) {
  unsigned int crate = MINOR(file->f_dentry->d_inode->i_rdev);
  int N=0;
  int i=0;
  char * temp;
  int *k;
  int which = crate-C111A_MAX_CRATES;
#ifdef _DEBUG
  printk("C111A DEBUG: Reading fifo %i\n", which);
#endif
  if (which >= FIFOS) return 0;
  temp = kmalloc(count, GFP_KERNEL);
  if (temp == NULL) {
    printk("C111A NOTIFY: Unable to kmalloc in read_fifo\n");
    return 0;
  }
  for (i=0; i<count; i++) {
    temp[i] = 0;
  }
#ifdef _DEBUG
  printk("C111A DEBUG: Allocated %i bytes\n", count);
#endif
  while ( last[which]==first[which] ) {
#ifdef _DEBUG
    printk("C111A DEBUG: No data in fifo. \n");
#endif
    sleep_on_counter++;
    INTERRUPTIBLE_SLEEP_ON(&wait_on_data[which]);
    if (signal_pending(current))
      return -ERESTARTSYS;
  }
  N = rtf_get(which, temp, count);
  k = (int *)temp;
#ifdef _DEBUG
  printk("C111A DEBUG: Read %i on %i (size is %i)\n", *k, which, N);
#endif
  if (N>0) {
    copy_to_user(buf, temp, count);
  } 
  kfree(temp);
  return N;
}

/*----------- IRQ functions ---------------------*/

static int IRQCrate=0;
static int IRQset=0;
static int NevtsToRead=0;
static int Nevts=0;
static int LAMMask[C111A_MAX_CRATES];

struct CAMAC_COMMAND {
  int CRATE;
  int STATION;
  int ADDRESS;
  int FUNCTION;
  int DTW;
  struct CAMAC_COMMAND * next;
};

static int N_CAMAC_Cmd = 0;
static int N_CAMAC_Words = 0;
struct CAMAC_COMMAND * List_Of_Commands;
static long nerrs=0;
int * dataBuffer;

static void irq_handler(int); 
int get_LAM_masked(int);

void * handler = &irq_handler;

static void irq_handler(int irq) {
  int i=0;
  int nwritten=-1;
  int crate=IRQCrate;
  int hb, lw;
  int count=0;
  int jret=0;
  int data=0;
  int ctrl=0;
  int n,a,f;
  struct CAMAC_COMMAND Command;
  int nWords = 0;
  int flags;

  mean_time_between_interrupts+=(jiffies-time_since_last_interrupt);
  time_since_last_interrupt = jiffies;
  interrupts++;
  
  save_flags(flags);
  cli();
  /* Check if the LAM is masked first     */
  /* Interrupt can only come from crate IRQCrate */
  outw_p((crate & 0x0F) | ((DELAY & 0x0F) << 0x08), GEO_port);  
  jret = 0;
#ifdef _DEBUG
  printk("C111A DEBUG: (handler) LAMMask=0x%06x\n",LAMMask[crate]);
#endif
  for (i=22; ((i>=0) & (jret==0)); i--) {
#ifdef _DEBUG
    printk("C111A DEBUG: (handler) Testing LAM on module %i (LAMMask=0x%06x)\n",
	   i+1,LAMMask[crate]);
    printk("C111A DEBUG: (handler) Bit %i = %i\n", i, 
	   ((LAMMask[crate] >> i) & 0x01));
#endif
    if (((LAMMask[crate] >> i) & 0x01) == 0x01) {
      outw_p(NAF2(i+1, 0, 8), FNA_port);
      count++;
#ifdef _DEBUG
      printk("C111A DEBUG: (handler) CAMAC Cmd called until now = %i\n",count);
#endif
      if ((inb_p(R_S) & 0x01)==0x01) {
	jret = i;
#ifdef _DEBUG
	printk("C111A DEBUG: (handler) Found LAM on module %i\n",
	   jret+1);
#endif
      }
    }
  }
  if (jret==0) { 
#ifdef _DEBUG
    printk("C111A DEBUG: (handler) No Masked LAM to serve...\n");
#endif
    restore_flags(flags);
    return; 
  } else {
#ifdef _DEBUG
    printk("C111A DEBUG: (handler) LAM to be served on module %i\n",jret+1);
    printk("C111A DEBUG: (handler) Starting executing commands...\n");
    printk("C111A DEBUG: (handler) Commands in queue: %i\n", N_CAMAC_Cmd);
#endif
    /* Starts data acquisition */
    if (List_Of_Commands == NULL) {
      /* We must at least reset the LAM */
      outw_p((crate & 0x0F) | ((DELAY & 0x0F) << 0x08), GEO_port);  
#ifdef _DEBUG
      printk("C111A DEBUG: (handler) Resetting LAM...\n");
#endif
      outw_p(NAF2(jret+1,0,9), FNA_port);
      restore_flags(flags);
      return;
    }
    Command = *List_Of_Commands;
    for (i=0; i<N_CAMAC_Cmd; i++) {
#ifdef _DEBUG
      printk("C111A DEBUG: (handler) Executing Command:"
	     "C=%2i N=%2i A=%2i F=%2i D=%2i\n",
	     Command.CRATE, Command.STATION, Command.ADDRESS, Command.FUNCTION,
	     Command.DTW);
#endif
      outw_p((Command.CRATE & 0x0F) | ((DELAY & 0x0F) << 0x08), GEO_port);  
      n = Command.STATION;
      a = Command.ADDRESS;
      f = Command.FUNCTION;
#ifdef _DEBUG
      printk("C111A DEBUG: (handler) CAMAC Cmd called until now = %i\n",count);
#endif
      data = 0;
      if (f<=7) {
	outw_p(NAF2(n,a,f), FNA_port);
	hb = inb_p(HB_R) & 0xFF; 
	lw = inw_p(LW_R) & 0xFFFF;
	data = (0xFFFF*(hb) + (lw)) ;

	ctrl = (inb_p(R_S) & 0x0f);
	/*  Pass data to the fifo  */
#ifdef _DEBUG
	printk("C111A DEBUG: (handler) Writing %i to FIFO\n", data);
	printk("C111A DEBUG: (handler) Ctrl %i N %i A %i F %i\n", ctrl, n, a, f);
#endif      
	dataBuffer[2*nWords] = data;
	dataBuffer[2*nWords+1] = ctrl+n*100+a*10000+f*1000000;
	nWords++;
      } else if (f>=16 && f<=23) {
	outw_p(NAF2(n,a,f)-0x8000, FNA_port);
	data = Command.DTW;
	outb_p(((data >> 0x10) & 0xFF),  HB_W);
	outw_p(data & 0xFFFF, LW_W);
	ctrl = (inb_p(R_S) & 0x0f);
      } else if ((f>=8 && f<=15) || (f>=24 && f<=31))  {
	outw_p(NAF2(n,a,f), FNA_port);
	ctrl = (inb_p(R_S) & 0x0f);
      } else {
	data = 0;
	ctrl = 0;
      }
#ifdef _DEBUG
      printk("C111A DEBUG: (handler) DATA=%i CTRL=%i\n", data, ctrl);
#endif
      /* next command */
      if (Command.next != NULL) Command = *(Command.next);
    }
    nwritten = rtf_put(DATAFIFO, dataBuffer, N_CAMAC_Words*2);
    if (nwritten!=N_CAMAC_Words*2) {
#ifdef _DEBUG
      printk("C111A ERROR: (handler) rtf_put return %i\n", nwritten);
#endif
      nerrs++;
      restore_flags(flags);
      return;
    }
    Nevts++;
#ifdef _DEBUG
    printk("C111A DEBUG: (handler) Events Served: %i/%i\n", Nevts,NevtsToRead);
#endif      
    if (Nevts>=NevtsToRead) {
      printk("C111A MSG  : (handler) mean time between interrupts: %d\n",
	     mean_time_between_interrupts/interrupts);
      printk("C111A MSG  :           interrupts: %d\n", interrupts);
      printk("C111A MSG  :           zero readings: %d     empty fifo: %d\n",
	     zero_readings, empty_fifo);
      printk("C111A MSG  :           sleep on: %d          wake_up: %d\n",
	     sleep_on_counter, wake_up);
      printk("C111A MSG  :           cyclings: %d\n",
	     cycling);
      zero_readings = 0;
      empty_fifo = 0;
      time_since_last_interrupt = jiffies;
      mean_time_between_interrupts = 0;
      interrupts = 0;
      sleep_on_counter = 0;
      wake_up = 0;

      /* Disable Interrupts                                */
      /* write Nevts on control fifo                      */
      /* write 2*N_CAMAC_Words*0xFFFF on data fifo */
      for (i=0; i<N_CAMAC_Words; i++) {
	dataBuffer[2*i]   = 0xFFFF;
	dataBuffer[2*i+1] = 0xFFFF;
      }
      nwritten = rtf_put(DATAFIFO, dataBuffer, N_CAMAC_Words*2);
      if (nwritten!=N_CAMAC_Words*2) {
#ifdef _DEBUG
	printk("C111A ERROR: (handler) rtf_put return %i at end\n",nwritten);
#endif
	nerrs++;
	Nevts--;
      } else {
	printk("C111A MSG  : (handler) Freeing IRQ %i after %i events\n",IRQ, 
	       Nevts);
	printk("C111A MSG  : (handler) Errors so far: %li\n", nerrs);
	//
	// disable LAM detect
	//
	outb_p(0x0D, S_mod_W);
	dataBuffer[0] = Nevts;
	printk("C111A MSG  : (handler) Putting %i into CTRL fifo\n",
	       dataBuffer[0]);
	rtf_put(CTRLFIFO, dataBuffer, 1);
	free_irq(IRQ,NULL);
	IRQset = 0;
	restore_flags(flags);
	return;
      }
    }
  }
  restore_flags(flags);
  return; 
}

int get_LAM_masked(int crate) {
  int istat = 0;
  int i;

  c111a_set(crate);
  for (i=0; ((i<=24) & ((LAMMask[crate] >> i)>0)); i++) {
#ifdef _DEBUG
    printk("C111A DEBUG: (get_LAM_masked) LAM Mask is 0x%06x\n",LAMMask[crate]);
    printk("C111A DEBUG: (get_LAM_masked) Testing LAM on module %i : ",i+1);
    printk("C111A DEBUG: (get_LAM_masked) 0x%06x\n",(LAMMask[crate] >> i) & 0x01);
    printk("C111A DEBUG: (get_LAM_masked) LAMMask complement:  0x%06x\n",(LAMMask[crate] >> (i+1)));
#endif
    if (((LAMMask[crate] >> i) & 0x01) == 0x01) {
      int nafLAM = NAF(i+1, 0, 8);
      c111a_load_naf(nafLAM);
      if (Q(c111a_status())==0x01) return i;
    }
  }
  return istat;
}

/*----------- board functions -------------*/

static void c111a_load_naf(int naf)
{
  int n=N_NAF(naf);
  int a=A_NAF(naf);
  int f=F_NAF(naf);
  naf = NAF2(n,a,f);
  /* if we have a write function then do not put S=1 into FNA_port! */
  if (f>=16 && f <=23) { 
    naf = naf-0x8000;
    outw_p(naf, FNA_port);
  } else {
#ifdef _DEBUG
    printk("C111A DEBUG: (c111a_load_naf) Loading FNA_port with 0x%4x\n",naf);
    printk("C111A DEBUG: (c111a_load_naf) Decoding: %i %i %i\n",N_NAF2(naf),
	   A_NAF2(naf), F_NAF2(naf));
#endif
    outw_p(naf, FNA_port);
  }
}

static int c111a_set(unsigned int crate) 
{
  /* set the GEO port and the crate */
  outw_p((crate & 0x0F) | ((DELAY & 0x0F) << 0x08), GEO_port);  
  outb_p(0x0F, Crate_on);
  if (currentDevice != crate) currentDevice = crate;
  return currentDevice;
}

static int c111a_test()
{
  /* test pc board: 0 = not found */
  int test = !(((inb_p(S_mod_R) & 0x04) >> 2) & 0x01);
#ifdef _DEBUG
  if (test==0) {
    printk("(C111A)[c111a_test]  Crate status is 0x%x\n", 
	   inb_p(S_mod_R));
  }
#endif
  return test;
}

static int c111a_status()
{
#ifdef _DEBUG
  printk("C111A DEBUG: (c111a_status)  Camac status is 0x%x\n",
	 inb_p(R_S) & 0x0f);
#endif  
  return (inb_p(R_S) & 0x0f);
}

static int c111a_read_24(void)
{
  unsigned int hb = inb_p(HB_R) & 0xFF; 
  unsigned int lw = inw_p(LW_R) & 0xFFFF;
#ifdef _DEBUG
  printk("C111A DEBUG: c111a_read_24) LW is %i, HB is %i\n",lw,hb);
#endif
  return (0xFFFF*(hb) + (lw)) ;
}

static void c111a_write_24(int data)
{
#ifdef _DEBUG
  printk("C111A DEBUG: (c111a_write_24)  LW is %i, HB is %i\n", 
	 data & 0xFFFF, ((data >> 0x10) & 0xFF));
#endif  
  outb_p(((data >> 0x10) & 0xFF),  HB_W);
  outw_p(data & 0xFFFF, LW_W);
}

/*----------- File System functions -------------*/

static long long c111a_lseek(struct file *file, long long offset, 
			       int orig) {
#ifdef _DEBUG
  printk("C111A DEBUG: (lseek_c111a) COMMAND: %i %i %i\n",
	 N_NAF(file->f_pos),A_NAF(file->f_pos),F_NAF(file->f_pos));
#endif
  if (orig == 0) {
    file->f_pos = offset;
    return file->f_pos;
  }
  else
    return -EINVAL;
}

static ssize_t c111a_write(struct file * file, const char * buf,
		        size_t count, loff_t *ppos)
{
  unsigned int crate = MINOR(file->f_dentry->d_inode->i_rdev);
  int result;
  int temp;

#ifdef _DEBUG
  printk("C111A DEBUG: (write) Requested function %i on crate %i\n",
         F_NAF(file->f_pos),crate);
#endif
  if (crate >= C111A_MAX_CRATES)
    return -ENODEV;

  /*
   * Check that we can really read from the buffer
   */

  if ((result = verify_area(VERIFY_READ,buf,count)))
    return result;
  c111a_set(crate);
  if (c111a_test()==0) {
    return -1;
  } else {
    memcpy(&temp, buf, count);
    c111a_load_naf(file->f_pos);
#ifdef _DEBUG
    printk("C111A DEBUG: (write) Writing %i as data\n",temp);
#endif
    c111a_write_24(temp);
    return c111a_status();
  }
}

static ssize_t c111a_read(struct file * file, char * buf,
		       size_t length, loff_t *ppos)
{
  unsigned int crate = MINOR(file->f_dentry->d_inode->i_rdev);
  int result;
  int temp;

#ifdef _DEBUG
  printk("C111A DEBUG: (read_c111a) Requested function %i on crate %i\n",
         F_NAF(file->f_pos),crate);
#endif
  if (crate >= C111A_MAX_CRATES) {
    return read_fifo(file, buf, length, ppos);
  }

  /*
   * Check that we can really write to the buffer
   */

  if ((result = verify_area(VERIFY_WRITE,buf,length)) && (buf != NULL))
    return result;

  /* Load N,A,F then return the number of bytes read */

  c111a_set(crate);
  if (c111a_test()==0) {
    return -1;
  }else {
    c111a_load_naf(file->f_pos);
    temp = c111a_read_24();
#ifdef _DEBUG
    printk("C111A DEBUG: (read_c111a) Read data is %i\n",temp);
#endif
    if (buf != NULL) {
      copy_to_user(buf, &temp, 4);
#ifdef _DEBUG
      printk("C111A DEBUG: (read_c111a) Buffer copied...\n");
#endif
    }
    return c111a_status();
  }
}

static int c111a_open(struct inode * inode, struct file * file)
{
  unsigned int crate = MINOR(inode->i_rdev);
  int i;
  int iret=0;

#ifdef _DEBUG
  printk("C111A DEBUG: (open_c111a) Requesting crate %i\n",crate);
#endif
  if (crate >= (C111A_MAX_CRATES+FIFOS)) {
    printk("C111A MSG: no such crate\n");
    return -ENODEV;
  }

  if (nDeviceOpen[crate] != 0) {
    printk("C111A MSG: no such crate available\n");
    return -EBUSY;
  }

  if (crate >= C111A_MAX_CRATES) {
    printk("C111A MSG  : (open_c111a) FIFO %i (%i) ready\n",
	   crate-C111A_MAX_CRATES, crate);
  } else {
    N_CAMAC_Cmd = 0;
    N_CAMAC_Words = 0;
    for (i=0; i<C111A_MAX_CRATES; i++) {
      LAMMask[i] = 0;
    }

    c111a_set(crate);
    /*
     * if the crate is off try first to switch it on
     *
     */
    if (c111a_test()==0) {
      outb_p(0x0F, Crate_on);
    }
    if (c111a_test()==0) {
      printk("C111A MSG  : Crate is off\n");
      return -1;
    } else {
      /*
       * Generate a Z,C cycle and disable LAM servicing by default 
       */
      outb_p(0x0F, Crate_on);
      outb_p(0x0F, Crate_Z);
      outb_p(0x0F, Crate_C);
      outb_p(0x0B, S_mod_W);
      printk("C111A MSG  : Crate %i opened and cleared\n",crate);
      IRQset = 0;
      /* Open RT fifoes                */
      /* 0: control FIFO               */
      /* 1: output FIFO                */
      rtf_destroy(0);
      rtf_destroy(1);
      iret = rtf_create(0);
#ifdef _DEBUG
      printk("C111A DEBUG: (open_c111a) FIFO 0 open result: %i\n", 
	     iret);
#endif
      iret = rtf_create(1);
#ifdef _DEBUG
      printk("C111A DEBUG: (open_c111a) FIFO 1 open result: %i\n", 
	     iret);
#endif
    }
  }
  nDeviceOpen[crate] = 1;
  return iret;  
}

static int c111a_ioctl(struct inode *inode, struct file *file,
		    unsigned int cmd, unsigned long arg)
{

  unsigned int crate = MINOR(inode->i_rdev);
  int buf[2];
  int f;
  int istat;
  int flags;
  struct CAMAC_COMMAND * aCommand, *currentCommand;

  f = F_NAF(file->f_pos);

#ifdef _DEBUG
  printk("C111A DEBUG: (ioctl_c111a) Requested function %i on crate %i\n",
          f, crate);
#endif
  if (crate >= C111A_MAX_CRATES) return -ENODEV;
  if (f < 32) {
#ifdef _DEBUG
    printk("C111A DEBUG: (ioctl_c111a) Standard Function\n");
#endif
    c111a_set(crate);
    if (c111a_test()==0) {
      return -1;
    } else {
      c111a_load_naf(file->f_pos);
      return c111a_status();
    }
  } else {
#ifdef _DEBUG
    if (f>=63) {
      printk("C111A DEBUG: (ioctl_c111a) STORE FUNCTION\n");
    } else {
      printk("C111A DEBUG: (ioctl_c111a) Non standard function\n");
    }
#endif
    if (c111a_test()==0) {
      return -1;
    } else {
      switch (f) {
      case 32: 
	/* Clear I */
	outb_p(0x0B, S_mod_W);
	break;
      case 33: 
	/* Set I */
	outb_p(0x0A, S_mod_W);
	break;
      case 34:
	/* Clear I, Z cycle */
	outb_p(0x0B, S_mod_W);
	outb_p(0x0F, Crate_Z);
	break;
      case 35:
	/* Set I, Z cycle */
	outb_p(0x0A, S_mod_W);
	outb_p(0x0F, Crate_Z);
	break;
      case 36:
	/* Clear I, C cycle */
	outb_p(0x0B, S_mod_W);
	outb_p(0x0F, Crate_C);
	break;
      case 37:
	/* Set I, C cycle */
	outb_p(0x0A, S_mod_W);
	outb_p(0x0F, Crate_C);
	break;
      case 38:
	/* Clear I, Z, C cycle */
	outb_p(0x0B, S_mod_W);
	outb_p(0x0F, Crate_Z);
	outb_p(0x0F, Crate_C);
	break;
      case 39:
	/* Set I, Z, C cycle */
	outb_p(0x0A, S_mod_W);
	outb_p(0x0F, Crate_Z);
	outb_p(0x0F, Crate_C);
	break;
      case 40:
	/* Disable Interrupts */
	if (IRQset==1) {
	  printk("C111A MSG  : (ioctl_c111a) Freeing IRQ %i. Events: %i\n",
		 IRQ, Nevts);
	  printk("C111A MSG  : (ioctl_c111a) Still %i events in queue\n",
		 last[1]-first[1]);
	  /* reset buffer */
	  first[1]=0;
	  last[1]=0;
	  cycling = 0;
	  free_irq(IRQ,NULL);
	  IRQset = 0;
	}
	//
	// disable LAM detect
	//
	outb_p(0x0D, S_mod_W);
	break;
      case 41:
	/* Enable Interrupts */
#ifdef _DEBUG
	printk("C111A DEBUG: (ioctl_c111a) Current statregs are 0x%x 0x%x\n",
	       inb_p(R_S) & 0x0f, inb_p(S_mod_R) & 0x0f);
#endif
	printk("C111A MSG  : (ioctl_c111a) Activating IRQ %i\n",IRQ);
	printk("C111A MSG  :               Data to pass to fifo: %i\n",
	       N_CAMAC_Words);
        if (IRQset==0) {
	  istat=request_irq(IRQ,handler,SA_INTERRUPT,"c111a",NULL);
	  nerrs = 0;
	} else {
	  istat = 0;
	}
	if (istat<0) {
	  save_flags(flags);
	  printk("C111A MSG  : (ioctl_c111a) request_irq returned %i. Flags=%i\n",
		 istat,flags);
	} else {
	  dataBuffer=(int *)kmalloc(N_CAMAC_Words*2*sizeof(int), GFP_KERNEL);
	  if (dataBuffer == NULL) {
	    printk("C111A NOTIFY: Unable to kmalloc in ioctl\n");
	    return 0;
	  }
	  IRQset   = 1;
	  IRQCrate = MINOR(inode->i_rdev);
	  save_flags(flags);
#ifdef _DEBUG
	  printk("C111A DEBUG: (ioctl_c111a) request_irq returned %i. Flags=%i\n",
		 istat,flags);
	  printk("C111A DEBUG: (ioctl_c111a) LAM generator crate : %i\n",
		 IRQCrate);	
#endif
	  Nevts   = 0;
	  //
	  // enable LAM detect
	  //
	  outb_p(0x0C, S_mod_W);
#ifdef _DEBUG
	  printk("C111A DEBUG: (ioctl_c111a) Current statregs are 0x%x 0x%x\n",
 		 inb_p(R_S) & 0x0f, inb_p(S_mod_R) & 0x0f);
#endif
	}
	break;
      case 42:
	/* Store LAM mask */
	printk("C111A MSG  : (ioctl_c111a) Set LAM Mask: 0x%06x\n",cmd);
	LAMMask[crate]=cmd;
	break;
      case 43:
	/* Clear IRQ handler */
	printk("C111A MSG  : (ioctl_c111a) Clearing irq handler\n");
	while (List_Of_Commands != NULL) {
	  aCommand = List_Of_Commands;
	  List_Of_Commands = aCommand->next;
	  kfree(aCommand);
	}
	if (dataBuffer!=NULL) {
	  kfree(dataBuffer);
	  dataBuffer = NULL;
	}
        N_CAMAC_Cmd = 0;
        N_CAMAC_Words = 0;
	break;
      case 48:
	/* Read LAM mask */
	return LAMMask[crate];
	break;
      case 56:
	/* Read LAM pending */
	istat = get_LAM_masked(crate);
	return istat;
	break;
      case 60:
	/* Set max no. of events in IRQ handler */
	/* Each time we set a new number of events reset the buffer */

	NevtsToRead = cmd;
	printk("C111A MSG  : (ioctl_c111a) No. of evts to read: %i\n", 
	       NevtsToRead);
	return NevtsToRead;
	break;
      case 61:
	/* write 0xFF on data fifo and ctrl fifo (to be used by users */
	/* to test the end of a DAQ cycle)                            */
	buf[0] = 0xFFFF;
	buf[1] = 0xFFFF;
	rtf_put(CTRLFIFO, &buf[0], 1);
	rtf_put(DATAFIFO, &buf[0], 1);
	break;
      default:
	/*
         * we accept the NAF code coming from applications and use it
         * to load the List_Of_Commands to be executed after irq.
         * To detect that the user wants to load List_Of_Commands instead
         * of execute them, we use CAMAC function codes + 100. Then
         * we load the CAMAC_COMMAND structure with f-100.
         *
         */
	aCommand = (struct CAMAC_COMMAND *)kmalloc(sizeof(struct CAMAC_COMMAND), GFP_DMA);
	if (aCommand == NULL) {
	  printk("C111A NOTIFY: Unable to kmalloc in ioctl (def cmd)\n");
	  return 0;
	}
	aCommand->CRATE    = crate;
	aCommand->STATION  = N_NAF(file->f_pos);
	aCommand->ADDRESS  = A_NAF(file->f_pos);
	aCommand->FUNCTION = F_NAF(file->f_pos)-100;
	aCommand->DTW      = cmd;
	aCommand->next     = NULL;
	N_CAMAC_Cmd++;
	if (aCommand->FUNCTION <= 7) {
	  N_CAMAC_Words++;
	}
	printk("C111A MSG  : (ioctl_c111a) COMMAND:%2i %2i %2i %2i %2i\n",
	       aCommand->CRATE, aCommand->STATION, aCommand->ADDRESS,
	       aCommand->FUNCTION, aCommand->DTW);
#ifdef _DEBUG
	printk("C111A DEBUG: (ioctl_c111a) Commands stored : %i\n",N_CAMAC_Cmd);
#endif
	if (List_Of_Commands == NULL) {
	  List_Of_Commands = aCommand;
	} else {
	  currentCommand = List_Of_Commands;
	  while (currentCommand->next != NULL) {
	    currentCommand = currentCommand->next; 
	  }
	  currentCommand->next = aCommand;
	} 
      }
      istat = inb_p(S_mod_R) & 0x0f;
#ifdef _DEBUG
      printk("C111A DEBUG: (ioctl_c111a) Returning status 0x%x\n",istat);
#endif
      return (istat & 0x07);
    }
  }
}

static int c111a_release(struct inode *inode, struct file *file)
{
  unsigned int crate = 0;
  struct CAMAC_COMMAND * Command;
  
  crate =  MINOR(file->f_dentry->d_inode->i_rdev);
  printk("C111A MSG  : (close_c111a) Closing crate %i\n", crate);
    
  if (crate >= (C111A_MAX_CRATES+FIFOS)) {
    printk("C111A MSG  : (close_c111a) No such crate to close\n");
    return -EIO;
  }
  
  if (nDeviceOpen[crate]==0) {
    printk("C111A MSG  : (close_c111a) No such opened crate to close\n");
    return 0;
  }
  
  nDeviceOpen[crate] = 0;

  if (crate < C111A_MAX_CRATES) {
    c111a_set(crate);
    if (c111a_test()) {
      outb_p(0x0F, Crate_off);
#ifdef _DEBUG
      printk("C111A DEBUG: (close_c111a) Crate %i powered off\n",crate);
#endif
    }
  }

  if (crate >= C111A_MAX_CRATES) {
      printk("C111A MSG  : (close_c111a) Fifo %i closed\n",
	     crate-C111A_MAX_CRATES);
      rtf_destroy(crate-C111A_MAX_CRATES);
#if LINUX_VERSION_CODE < LinuxVersionCode(2,4,0)
      wait_on_data[crate-C111A_MAX_CRATES] = NULL;
#else
      init_waitqueue_head(&wait_on_data[crate-C111A_MAX_CRATES]);
#endif
      return 0;
  }
  /* free the interrupt handler (if any) */
  if (IRQset==1) {
    printk("C111A MSG  : (close_c111a) Freeing IRQ %i. Events %i\n",
	   IRQ,Nevts);
    printk("C111A MSG  : (ioctl_c111a) Still %i events in queue\n",
	   last[1]-first[1]);
    /* reset buffer */
    first[1]=0;
    last[1]=0;
    cycling = 0;
    free_irq(IRQ,NULL);
    IRQset=0;
  }
  /* release allocated memory */
  while (List_Of_Commands != NULL) {
    Command = List_Of_Commands;
#ifdef _DEBUG
    printk("C111A DEBUG: (close_c111a) Freeing IRQ Command %2i %2i %2i\n",
	   Command->STATION, Command->ADDRESS, Command->FUNCTION);
#endif
    List_Of_Commands = Command->next;
    kfree(Command);
  }
  if (dataBuffer!=NULL) {
    kfree(dataBuffer);
    dataBuffer = NULL;
  }
  return 0;
}

#if LINUX_VERSION_CODE < LinuxVersionCode(2,4,0)
static struct file_operations c111a_fops = {
	c111a_lseek,
	c111a_read,
	c111a_write,
	NULL,		/* lp_readdir */
	NULL,		/* lp_poll */
	c111a_ioctl,
	NULL,		/* lp_mmap */
	c111a_open,
	NULL,		/* flush */
	c111a_release
};
#else
static struct file_operations c111a_fops = {
  llseek : c111a_lseek,  
  read   : c111a_read,  
  write  : c111a_write, 
  ioctl  : c111a_ioctl,  
  open   : c111a_open,
  release: c111a_release,
};
#endif


/* --- initialisation code ------------------------------------- */

int init_module(void)
{ 
  int count = 0;
  int crate;

  printk("===================================================================\n");
  printk("(C111A) Module C111A starting...\n");

  if (register_chrdev(C111A_MAJOR, C111A_NAME, &c111a_fops)) 
  {
    printk(" unable to get major %i for " C111A_NAME " device\n",major);
    return -EIO;
  }

  printk(STARTUPSTRING);
 
  zero_readings = 0;
  empty_fifo = 0;
  time_since_last_interrupt = jiffies;
  mean_time_between_interrupts = 0;
  interrupts = 0;
  sleep_on_counter = 0;
  wake_up = 0;

/*--------- forward declarations --------------*/

  if (check_region(GEO_port, C111A_BASE_EXTEND)) 
    {
      printk(" unable to access i/o address 0x%lx\n",GEO_port);
      unregister_chrdev(major, C111A_NAME);
      return -EIO;
    }

  c111a_set(0);
  if (!c111a_test()) 
    {
      printk(" interface card not found at address 0x%lx\n",GEO_port);
      //      unregister_chrdev(major, C111A_NAME);
      //      return -EIO;
    } else {
      printk(" interface card found at address 0x%lx\n",GEO_port);
    }
  request_region(GEO_port, C111A_BASE_EXTEND, C111A_NAME);
  
  currentDevice = -1;
 
  for (crate = C111A_MAX_CRATES; crate < C111A_MAX_CRATES+FIFOS; crate++) {
      nDeviceOpen[crate] = 0;
  }

  for (crate = 0; crate < C111A_MAX_CRATES; crate++) 
    {
      
      nDeviceOpen[crate] = 0;
      
      c111a_set(crate);
      if (c111a_test()==1) {
	count++;
	printk(" crate %3i [ ON]\n", crate);
      } else {
	printk(" crate %3i [OFF]\n", crate);
      }
    }

  for (crate = 0; crate < FIFOS; crate++) {
#if LINUX_VERSION_CODE < LinuxVersionCode(2,4,0)
    wait_on_data[crate] = NULL;
#else
    init_waitqueue_head(&wait_on_data[crate]);
#endif

  }
  if (count == 0)
    printk(" no crate found\n");

  printk("===================================================================\n");
  return 0;
}

void cleanup_module(void)
{
  if (MOD_IN_USE) {
    printk("(C111A) device busy, module not removed\n");
    return;
  }
  
  printk(" Realeasing crate at address: 0x%lx\n",GEO_port);
  release_region(GEO_port, C111A_BASE_EXTEND);

  if (unregister_chrdev(major, C111A_NAME) != 0)
    printk("(C111A) cleanup_module failed\n");
  else
    printk("(C111A) module removed\n");
  
}




